#!/bin/bash

usage() {
    cat <<EOF
Usage: $0 [options]

Options:
   -h                   show this help message and exit
   -r RELEASEVER        Value to use for \$releasever (example: epel-5,
                        fedora-14, fedora-rawhide)
   -a ARCH              Value used for \$basearch
   -d RESULTDIR         Directory to store results (default: \$PWD)
   -e NVR               One or more RPM NVR's to download from koji
   -t TMPDIR            Temporary directory to use (default: /tmp)
EOF
    exit 1
}

## Detect the system release and version
detect_releasever() {
    RELEASE="fedora"
    VER="rawhide"
    # Guess releasever based on current system release
    if [ -f /etc/fedora-release ]; then
        RELEASE="fedora"
        VER=$(gawk '{print $3}' /etc/fedora-release)
    elif [ -f /etc/redhat-release -a ! -L /etc/redhat-release ]; then
        RELEASE="epel"
        VER=$(gawk '{print $5}' /etc/redhat-release)
        case $VER in
            4* ) VER=4 ;;
            5* ) VER=5 ;;
            6* ) VER=6 ;;
            * ) VER=6 ;;
        esac
    fi
    echo "$RELEASE-$VER"
}

error() {
    echo "Error: $1"
    exit 1
}

sigterm() {
    cleanup
    exit 1
}

cleanup() {
    test -n "${SETUP}" && rm ${SETUP}
    rm -f ${MOCKDIR}/${MOCKCFG}.cfg
}

# Default values
BASEARCH=$(uname -i)
RELEASEVER=$(detect_releasever)
RESULTSDIR=${PWD}/results/$(date +%Y%m%d)
TMPDIR=/tmp
NVR=""
MOCKDIR=/etc/mock

# Process arguments
while getopts "ha:r:d:t:e:" option
do
  case ${option} in
    a ) BASEARCH=${OPTARG} ;;
    r ) RELEASEVER=${OPTARG} ;;
    d ) RESULTSDIR=${OPTARG} ;;
    t ) TMPDIR=${OPTARG} ;;
    e ) NVR="${NVR} ${OPTARG}" ;;
    \?|h ) usage ;;
    * ) usage ;;
  esac
done

# Sanitize input arguments
RELEASE=${RELEASEVER%%-*}
VERSION=${RELEASEVER##*-}
test -d $MOCKDIR || error "Directory $MOCKDIR not found.  Is mock installed?"
# FIXME - sanitize basearch
# FIXME - sanitize releasever

# Prepare mock config template
MOCKROOT="${RELEASEVER}-${BASEARCH}"
MOCKCFG="composetree-${MOCKROOT}"
cat << EOF > ${MOCKDIR}/${MOCKCFG}.cfg
# Enable composes
config_opts['plugin_conf']['bind_mount_enable'] = True
config_opts['plugin_conf']['bind_mount_opts']['dirs'].append(('/dev','/dev/'))
config_opts['plugin_conf']['bind_mount_opts']['dirs'].append(('/dev/pts','/dev/pts/'))
config_opts['plugin_conf']['bind_mount_opts']['dirs'].append(('/dev/shm','/dev/shm/'))

EOF

cat ${MOCKDIR}/${MOCKROOT}.cfg >> ${MOCKDIR}/${MOCKCFG}.cfg
if [ "$?" -ne 0 ]; then
    error "Failed while building mock configuration file"
fi

# Install new packages for testing
KSCFG=$(mktemp --suffix .ks ${TMPDIR}/compose_tree-ks-XXXXXX)
SETUP=$(mktemp --suffix .sh ${TMPDIR}/compose_tree-setup-XXXXXX)
cat > ${SETUP} << EOF
#!/bin/sh

KOJI_RPMS=\$(mktemp -d /tmp/koji-rpms.XXXXXX)
mkdir -p \${KOJI_RPMS}
cd \${KOJI_RPMS}

## Use koji to download updated packages
for E in ${NVR}
do
    koji download-build --arch noarch --arch ${BASEARCH} \$E
done

## Install updated packages
yum -y localupdate *.rpm

## Make sure the requested packages were installed
rpm -q ${NVR}
## FIXME - exit on failure

## Create repository
createrepo \${KOJI_RPMS}

# Use existing spin-kickstart file, but use a smaller %packages
# Remove everything between %packages ... %end
gawk '/^%packages/,/^%end/ {next} {print}' /usr/share/spin-kickstarts/fedora-install-fedora.ks > ${KSCFG}
# Add @core
cat >> ${KSCFG} << EOKS
%packages --nobase
@core
%end

# Add custom repo for local koji packages
repo --name=koji-rpms --baseurl=file://\${KOJI_RPMS}
EOKS
EOF

# Ensure sigterm is called on exit
trap sigterm SIGTERM

MOCK="mock -r ${MOCKCFG}"
let RETVAL=0
${MOCK} --init
let RETVAL+=$?
${MOCK} --install koji pungi spin-kickstarts createrepo
let RETVAL+=$?
${MOCK} --copyin ${SETUP} /tmp/setup.sh
let RETVAL+=$?
${MOCK} --chroot "bash /tmp/setup.sh"
let RETVAL+=$?
${MOCK} --chroot "pungi --force --nosource --nodebuginfo -G -C -B --force --flavor Fedora --name Fedora --ver ${VERSION} -c ${KSCFG}"
let RETVAL+=$?

# Report the results.  We do all the copies individually in case one of the files doesn't exist.
mkdir -p ${RESULTSDIR}
cp /var/lib/mock/${MOCKROOT}/root/${KSCFG} ${RESULTSDIR}/ks.cfg
let RETVAL+=$?
cp /var/lib/mock/${MOCKROOT}/result/root.log ${RESULTSDIR}
let RETVAL+=$?

# Save generated installer files
for FNAME in .treeinfo images/boot.iso images/pxeboot/vmlinuz images/pxeboot/initrd.img
do
    ${MOCK} --quiet --copyout /${VERSION}/Fedora/${BASEARCH}/os/${FNAME} ${RESULTSDIR}
    let RETVAL+=$?
done

# Clean up and exit
cleanup
exit ${RETVAL}
