#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

# Notice: Most recent documentation is available at doc/test_class.py.template.

import os, sys, re, urlparse, urlgrabber
import autoqa.util
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autotest_lib.client.bin import utils
from autoqa.repoinfo import repoinfo
from autoqa.bodhi_utils import bodhi_list, bodhi_post_testresult, bodhi_already_commented, query_update
from autoqa.koji_utils import SimpleKojiClientSession

### FIXME: AutoQA should provide testarch
### https://fedorahosted.org/autoqa/ticket/272
localarch = autoqa.util.get_basearch()
testarch = localarch

# TODO: move these into koji_utils
def list_pending_builds(repotag):
    '''List pending builds for the repo associated with the given koji tag'''
    pending_tag = repotag + '-pending'
    koji = SimpleKojiClientSession()
    builds = koji.listTagged(pending_tag)
    nvrs = sorted([build['nvr'] for build in builds])
    return nvrs

def strip_epoch(rpmstr):
    return rpmstr.split(':',1)[-1] # works for E:N-V-R.A or N-V-R.A
def strip_arch(rpmstr):
    return rpmstr.rsplit('.',1)[0]

# --- Everything below here is depcheck-specific

class depcheck(AutoQATest):
    version = 2 # increment this if setup() changes

    @ExceptionCatcher()
    def setup(self, *args, **kwargs):
        utils.system('yum -y install python-rpmfluff mash')

    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.result = 'PASSED'
        self.highlights = []
        # dict of problematic NVRs - can't find in Bodhi
        # nvr -> error
        self.problematic = dict()

    def is_accepted(self, update, arch=testarch):
        '''Return True if the given update has been marked as Accepted by a
        previous depcheck run, False otherwise.'''
        try:
            (result, time) = bodhi_already_commented(update, 'depcheck', arch)
            return result == 'PASSED'
        except ValueError as e:
            msg = 'ERROR: %s' % e
            print msg
            self.highlights.append(msg)
            self.problematic[update] = e
            return False

    @ExceptionCatcher()
    def run_once(self, nvrs, kojitag, **kwargs):
        super(self.__class__, self).run_once()

        # Get our inputs
        pending = set(list_pending_builds(kojitag))
        # XXX set testarch to noarch for noarch updates?
        accepted_builds_nvrs = set([nvr for nvr in pending if self.is_accepted(nvr)])
        print 'Accepted builds: %s' % sorted(accepted_builds_nvrs)
        pending_builds_nvrs = sorted(pending.difference(accepted_builds_nvrs).difference(self.problematic.keys()))
        print 'Pending builds: %s' % pending_builds_nvrs
        print 'Problematic builds: %s' % self.problematic.keys()

        # Fetch packages and build commandline
        repo = repoinfo.getrepo_by_tag(kojitag)
        repoid = repo['name']
        cmd = './depcheck %s' % repoid
        for (nvr, rpms) in self.fetch_nvrs(accepted_builds_nvrs, self.tmpdir):
            for rpm in rpms:
                cmd += ' -a %s' % rpm

        # pending_parent_map keys() is a set of all downloaded rpms (stripped from
        # /path/OpenThreads-2.8.3-10.fc15.i686.rpm to OpenThreads-2.8.3-10.fc15).
        # Value for the key is then nvr of the build containing this rpm file.
        #
        # Example:
        # --------
        # Fetching RPMs for OpenSceneGraph-2.8.3-10.fc15
        #    OpenSceneGraph-2.8.3-10.fc15.i686.rpm
        #    OpenThreads-2.8.3-10.fc15.i686.rpm
        #
        # pending_parent_map = {OpenSceneGraph-2.8.3-10.fc15 : OpenSceneGraph-2.8.3-10.fc15,
        #                       OpenThreads-2.8.3-10.fc15 : OpenSceneGraph-2.8.3-10.fc15}
        pending_parent_map = {}
        for (nvr, rpms) in self.fetch_nvrs(pending_builds_nvrs, self.tmpdir):
            for rpm in rpms:
                cmd += ' %s' % rpm
                # While we're here, construct subpackage NVR -> parent NVR map
                # package filenames are all: /some/path/to/N-V-R.A.rpm
                this_nvr = os.path.basename(strip_arch(rpm[:-4]))
                pending_parent_map[this_nvr] = nvr

        # Run the test
        self.outputs = utils.system_output(cmd, retain_output=True)

        # variable 'results' is dict containing three keys IGNORE, REJECT, ACCEPT
        # value for each of the keys is list of rpm files
        # ignored/rejected/acceped by the depcheck test.
        #
        # Note that the '.rpm' is removed from the file name.
        # The dict then could be e.g.
        #
        # results = {'ACCEPT': [OpenSceneGraph-2.8.3-10.fc15.i686],
        #            'REJECT': [OpenThreads-2.8.3-10.fc15.i686]}

        # Process output
        results = dict(IGNORE=[], REJECT=[], ACCEPT=[])
        data_re = re.compile('^(IGNORE|REJECT|ACCEPT): (.*)$')
        for line in self.outputs.splitlines():
            match = data_re.match(line)
            if not match:
                continue
            (what, builds) = match.groups()
            results[what] = builds.split()

        # We need to split ENVRS to the groups according to update titles
        # since one update may contain more envrs, and we want to report
        # results for updates, not envrs
        def envra_to_nvr(envra):
            return strip_arch(strip_epoch(envra))


        # sets accepted_builds, rejected_builds, ignored_builds map the
        # rpm filenames to the build names which contain the respective
        # rpm file (taken from the pending_parent_map).
        # One example better than thousand words:
        #
        # results = {'ACCEPT': [OpenSceneGraph-2.8.3-10.fc15.i686],
        #            'REJECT': [OpenThreads-2.8.3-10.fc15.i686]}
        #
        # accepted_builds = set([OpenSceneGraph-2.8.3-10.fc15])
        # rejected_builds = set([OpenSceneGraph-2.8.3-10.fc15])
        #
        # (because both rpms are part of the OpenSceneGraph-2.8.3-10.fc15
        # build).

        accepted_builds = set()
        rejected_builds = set()
        ignored_builds = set()
        for (set_, keyword) in [(accepted_builds, 'ACCEPT'),
                                 (rejected_builds, 'REJECT'),
                                 (ignored_builds, 'IGNORE')]:
            for envra in results[keyword]:
                nvr = envra_to_nvr(envra)
                if pending_parent_map.has_key(nvr):
                    set_.add(pending_parent_map[nvr])
                else:
                    print "Build in %s but not in pending updates, skipping: %s" % (keyword, envra)

        # Now we need inspect accepted_builds, and discard the builds which
        # are also present in the rejected_builds (or ignored_builds) set.
        #
        # This is because depcheck outputs accepted/rejected rpms, which
        # we then map to builds. And if you look at the previous examples
        # it's apparent that since one rpm from the build passed, and
        # the second failed, the build's nvr is in both sets.
        #
        # But the build should pass only if all the rpms passed.
        #
        # After this step, accepted_builds and rejected_builds should be:
        #
        # accepted_builds = set([])
        # rejected_builds = set([OpenSceneGraph-2.8.3-10.fc15])

        accepted_builds = accepted_builds.difference(rejected_builds)
        accepted_builds = accepted_builds.difference(ignored_builds)

        # This code simply takes all builds, and creates a dictionary
        # of update metadata (taken from bodhi).
        # Because as one build can contain multiple rpms, one update
        # can contain multiple builds. This groups builds into updates.
        # The variable 'updates' then looks like
        #
        # updates = {updatename : bodhi_metadata}

        updates = {}
        all_builds = set(pending_parent_map.values())
        for nvr in all_builds:
            bodhi_info = None
            bodhi_info = query_update(nvr)
            if bodhi_info is None:
                print >> sys.stderr, "Update containing %s not found in Bodhi" % nvr
                continue

            if bodhi_info['title'] not in updates.keys():
                updates[bodhi_info['title']] = bodhi_info

        # Becase bodhi is working on top of updates, we need to check whether
        # all builds from the respective update were accepted. If at least
        # one of the builds in update was in rejected/ignored set, whole
        # update is 'failed'.
        #
        # If at least one update in the whole -pending repo is failed,
        # the overall depcheck result is FAILED.
        update_result = dict() # update_title -> update_result
        for update in updates.values():
            desc = 'Update %(result)s: %(title)s\n'
            update_result[update['title']] = 'PASSED'
            for build in update['builds']:
                if build['nvr'] not in accepted_builds:
                    update_result[update['title']] = 'FAILED'
                    self.result = 'FAILED'
                desc += self.build_highlights(build['nvr'], pending_parent_map,
                        accepted_builds, rejected_builds, ignored_builds)
            desc = desc % {'title': update['title'], 'result': update_result[update['title']]}
            desc = desc.rstrip() # strip last linefeed
            self.highlights.append(desc)
            print desc

        # Now let's sanity check depcheck output. Every pending build passed
        # to depcheck on command line must be either in accepted_builds, or
        # rejected builds, or ignored builds. If we passed a build to depcheck
        # and it didn't printed it out in the results, something is very wrong
        # and we want to cancel the whole test.

        missing_builds = [build for build in pending_builds_nvrs if
                                build not in accepted_builds and
                                build not in rejected_builds and
                                build not in ignored_builds]
        assert not missing_builds, 'Some builds passed in as pending, but ' \
                                   'depcheck produced no result for them: %s' % missing_builds

        # Report all results to Bodhi
        for title, result in update_result.items():
            print "Posting to Bodhi: %s %s" % (title, result)
            bodhi_post_testresult(title, 'depcheck', result,
                                  self.autotest_url, testarch, karma=0)

        if self.result == 'PASSED' and len(self.problematic) > 0:
            print >> sys.stderr, 'Problematic builds found, setting status to NEEDS_INSPECTION'
            self.result = 'NEEDS_INSPECTION'

        # TODO: force filling up reports for updates (resultsdb stuff)

        self.summary = kojitag

    def build_highlights(self, build_nvr, pending_parent_map, accepted_builds, rejected_builds, ignored_builds):
        build_rpms = set([key for key, value in pending_parent_map.items() if value == build_nvr])
        desc = '  Build %(result)s: %(build)s\n'

        rej = ""
        for rpm in build_rpms.intersection(rejected_builds):
            rej += '      %s\n' % rpm
        ign = ""
        for rpm in build_rpms.intersection(ignored_builds):
            ign += '      %s\n' % rpm

        if rej:
            desc += '    Ignored:\n'
            desc += rej
        if ign:
            desc += '    Rejected:\n'
            desc += ign

        result = 'FAILED' if build_nvr not in accepted_builds else 'PASSED'
        desc = desc % {'result': result, 'build': build_nvr}
        return desc

    def fetch_nvrs(self, nvrs, localdir):
        '''Fetch all the RPMs for the given package NVRs.
        Return a list of tuples: [(nvr,rpmlist), ...]'''
        # NOTE: Currently we're just fetching everything from koji. This might not
        # be the fastest thing, but it works for everyone, everywhere. If there's
        # a shortcut we can use on the autoqa test systems, we should use that,
        # but it needs to gracefully fall back to something like this.

        # TODO: Download only needed archs to speed it up
        # https://fedorahosted.org/pipermail/autoqa-devel/2011-February/001615.html
        # https://fedorahosted.org/autoqa/ticket/272
        rpms = []
        koji = SimpleKojiClientSession()
        for nvr in sorted(nvrs):
            pkgs = []
            # XXX deal with debuginfo/src RPMs?
            print "Fetching RPMs for %s" % nvr
            try:
                for rpm in koji.nvr_to_rpms(nvr, debuginfo=False, src=False):
                    urlobj = urlparse.urlparse(rpm['url'])
                    rpmname = os.path.basename(urlobj.path)
                    outpath = os.path.join(localdir, rpmname)
                    print "  %s" % rpmname
                    outfile = autoqa.util.grabber.urlgrab(rpm['url'], outpath)
                    pkgs.append(outfile)
            except urlgrabber.grabber.URLGrabError as e:
                self.result = 'ABORTED'
                raise
            print
            rpms.append((nvr, pkgs))
        return rpms
